/*******************************************************************
 *	Author:	Peter van Sebille (peter@yipton.demon.co.uk)
 *	Based on code by Otfried Cheong
 *
 *******************************************************************/


/*
 * Thanks to Ian Jeffray for the following info...

	I'm not sure if the following info is "public knowledge" but I
	know the following virtual addresses in ER5, which might help:

	Page Directory starts at 41000000
	Page Table Info array    41080000  
	Page Tables are at       42000000
	ROM is at                50000000
	Hardware mmap'd IO at    58000000
	Superpage / CPU page     40000000
	RAM Drive at             61000000

	the virt to phys map takes the top 12bits of the virtual address
	as an index in to the 16k page directory table (4096 words).

	The 32bit value which is selected is the PDE (Page Directory Entry).

	The PDE is a pointer to the page table : a 1K table of 32bit items.
	(Ie: One PDE maps 1MB of RAM)

	The next 8bits of the virtual address are an index in to the page
	table, giving the 32bit value which is the start of the physical page.
	(each page is 4K of course).

	The last 12bits of the virtual address select the byte in the
	physical page.


	Hope this isn't teaching folks to suck eggs, and may help?
*/

/*
 * Peter's comments...

	
	Initially I thought that I could find page table n at linear address:
	0x42000000 + n*1024. However that crashed machine for some/most linear
	addresses.
	After some tinkering it seems that page table n can be found by the following:
	Take the n-th TUint32 from base 0x41080000 (Page Table Info array).
	The 16 most significant bit provide the actual index in the Page Tables array.
*/

#include "tuiedit.h"

#include <e32std.h>
#include <e32cons.h>
#include <f32file.h>
#include <e32svr.h>

#include "arlo.h"


//#define DEBUG_PRINTS_PLEASE

#ifdef DEBUG_PRINTS_PLEASE
#define DBGPRINT(x) sConsole->Printf x
#define DBGPRESSKEY(x) sConsole->Printf x;sConsole->Getch()
#else
#define DBGPRINT(x) /**/
#define DBGPRESSKEY(x) /**/
#endif


#define BOOT_STRING_SIZE			64
#define PAGE_SIZE					4096
#define NR_PAGE_TABLE_ENTRIES		((PAGE_SIZE >> 2) - 2)


// globals
LOCAL_D CConsoleBase*		sConsole;
LOCAL_D RFs					sFs;
LOCAL_D CLineEdit*			sCommandLine;

LOCAL_D TUint32*			sArloPageTable;
LOCAL_D TInt				sArloPageTableIndex;
TText8 *					sBootString;

LOCAL_C TUint32 Linear2Physical(TUint32 aLinearAdr, RArLo& aArlo);

void ArloBoot(RArLo& aArlo)
{
	sConsole->Printf(_L("Booting kernel.... (press any key)\n"));
	sConsole->Getch();

	TUint32 arloPageTable = Linear2Physical((TUint32)sArloPageTable, aArlo);
	aArlo.Boot(arloPageTable, sArloPageTable[2]);		// and we're gone
}



LOCAL_C void LoadFileL(const TDesC& aFileName, RArLo& aArlo)
{
	RFile		file;

	sConsole->Printf(_L("Opening file %S\n"), &aFileName);
	User::LeaveIfError(file.Open(sFs, aFileName, EFileRead));

	TInt size;
	User::LeaveIfError(file.Size(size));
	sConsole->Printf(_L("File size %x\n"), size);

	TUint8 *buf = new (ELeave) TUint8[size + PAGE_SIZE];
	TUint32 loadAdr = TUint32(buf);
	loadAdr += PAGE_SIZE - 1;
	loadAdr &= ~(PAGE_SIZE - 1);

	buf = (TUint8 *) loadAdr;
	TPtr fileBuf(buf, size);
	sConsole->Printf(_L("Loading %x bytes to address %08x\n"), size, loadAdr);
	User::LeaveIfError(file.Read(fileBuf));
	sConsole->Printf(_L("Read %x bytes\n"), fileBuf.Length());
	if (fileBuf.Length() != size)
	{
		User::Leave(KErrNotFound);
	}
	file.Close();

	TInt	blockNo = (size + PAGE_SIZE - 1) / PAGE_SIZE;    // No of 4k blocks needed
	sConsole->Printf(_L("No pages = %d\n"), blockNo);

	if (sArloPageTableIndex + blockNo > NR_PAGE_TABLE_ENTRIES) 
	{
		sConsole->Printf(_L("Does not fit in page table.\n"));
		User::Leave(KErrGeneral);
	}
	// add entries to page map 
	for (TInt i = 0; i < blockNo; i++)
	{
		sArloPageTable[sArloPageTableIndex++] = Linear2Physical(loadAdr + i * PAGE_SIZE, aArlo);
	}
	sArloPageTable[sArloPageTableIndex++] = 0x0;
}

static void LoadFile(const TDesC& aFileName, RArLo& aArlo)
{
	TRAPD(err, LoadFileL(aFileName, aArlo));
	if (err)
	{
		sConsole->Printf(_L("Error loading file err=0x%x\n"), err);
	}
}

static TUint ArloExamine(RArLo& aArlo, TUint32 aAddressToExamine)
{
	TUint32		val = (TUint32) aAddressToExamine;

	DBGPRINT((_L("ArloExamine: peeking at 0x%x\n"), aAddressToExamine));
	DBGPRESSKEY((_L("Press a key\n")));

	aArlo.Examine(&val);

	DBGPRINT((_L("ArloExamine: value = 0x%x\n"), val));

	return (TUint) val;
}


static void SetupArloPageTableL(RArLo& aArlo)
{
	TText8*	table = (TText8*) User::AllocL(PAGE_SIZE * sizeof(TUint32) * 2);
	Mem::FillZ(table, PAGE_SIZE * sizeof(TUint32) * 2);
	table += (PAGE_SIZE -1);
	sArloPageTable = (TUint32*) (((TUint32) table) & ~(PAGE_SIZE - 1));

	sConsole->Printf(_L("sArloPageTable=0x%x\n"), sArloPageTable);

	sBootString = (TText8 *) new (ELeave) TUint32[BOOT_STRING_SIZE]; // word align 
	Mem::FillZ(sBootString, BOOT_STRING_SIZE * sizeof(TUint32));

	// first entry in page map:  phys address of option string
	TUint32 result = Linear2Physical((TUint32) sBootString, aArlo);

	sConsole->Printf(_L("Press a key\n"));
	sConsole->Getch();

	sArloPageTable[0] = result;

	sConsole->Printf(_L("Option buffer at phys address %08x\n"), sArloPageTable[0]);
	
	sConsole->Printf(_L("Press a key\n"));
	sConsole->Getch();

	// second entry: phys address of memory description
	TUint32		adr = 0x40000030;
#ifndef __WINS__
	adr = ArloExamine(aArlo, adr);
#endif
	sArloPageTable[1] = Linear2Physical(adr, aArlo);
	sConsole->Printf(_L("Memory description at address %08x\n"), sArloPageTable[1]);
	
	sArloPageTableIndex = 2;
}




static void CopyFileL(TDesC& aFileName)
{
	CFileMan*	fileman = CFileMan::NewL(sFs);
	User::LeaveIfError(fileman->Copy(aFileName, _L("c:\\"), ETrue));
	delete fileman;
}


static void ArloCopyFile(TDesC& aFileName)
{
	TRAPD(err, CopyFileL(aFileName));
	if (err)
	{
		sConsole->Printf(_L("Error copying file %S, err=%d\n"), &aFileName, err);
	}
	else
	{
		sConsole->Printf(_L("Successfully copied %S\n"), &aFileName);
	}
}


#if NOT_USED
static TUint32 Linear2Physical(TUint32 aLinearAdr, RArLo &arlo)
{
	TUint32 aPageDir = 0x40000014;

	aPageDir = ArloExamine(arlo, aPageDir);

	// Get virtual address of page directory
	sConsole->Printf(_L("Page directory %08x\n"), aPageDir);
	
	TUint32 aLevelOne = aPageDir + ((aLinearAdr & 0xfff00000) >> 18);
	sConsole->Printf(_L("Level One address %08x\n"), aLevelOne);
	
	aLevelOne = ArloExamine(arlo, aLevelOne);
	sConsole->Printf(_L("Level One descriptor %08x\n"), aLevelOne);
	
	// we now have the level one descriptor
	if ((aLevelOne & 0x3) == 0x2) 
	{
		sConsole->Printf(_L("Section\n"));
		// a Section descriptor
		return ((aLevelOne & 0xfff00000) | (aLinearAdr & 0x000fffff));
	}
	if ((aLevelOne & 0x3) != 0x1) 
	{
		sConsole->Printf(_L("fault or reserved\n"));
		// Fault or reserved
		return 0;
	}

	// Page descriptor.  Get Page table
	TUint32 aLevelTwo = (aLevelOne & 0xfffffc00);
	// convert Page table to linear
	sConsole->Printf(_L("physical aLevelTwo=0x%x\n"), aLevelTwo);
	aLevelTwo = RMmu::LinearAddress(aLevelTwo);
	sConsole->Printf(_L("lineair aLevelTwo=0x%x\n"), aLevelTwo);
	aLevelTwo += ((aLinearAdr & 0x000ff000) >> 10);

	sConsole->Printf(_L("Level Two address %08x\n"), aLevelTwo);
	aLevelTwo = ArloExamine(arlo, aLevelTwo);
	
	sConsole->Printf(_L("Level Two descriptor %08x\n"), aLevelTwo);
	// level two descriptor
	if ((aLevelTwo & 0x3) == 0x2) 
	{
		// a small (4K) page
		return ((aLevelTwo & 0xfffff000) | (aLinearAdr & 0x00000fff));
	}
	if ((aLevelTwo & 0x3) == 0x1) 
	{
		// a large (64K) page
		return ((aLevelTwo & 0xffff0000) | (aLinearAdr & 0x0000ffff));
	}
	
	// fault or reserved
	return 0;
}
#endif


#ifdef __WINS__

LOCAL_C TUint32 Linear2Physical(TUint32 aLinearAdr, RArLo& /*aArlo*/)
{
	return aLinearAdr;
}

#else

LOCAL_C TUint32 Linear2Physical(TUint32 aLinearAdr, RArLo& aArlo)
{
	TUint		pageDirectoryIndex = (aLinearAdr >> 20);
	TUint		pageTableIndex = ((aLinearAdr & 0x000ff000) >> 12);
	TUint		offsetInPage = aLinearAdr & 0xfff;

	DBGPRINT((_L("Virtual adr          =0x%08x\n"), aLinearAdr));
	DBGPRINT((_L("  pageDirectoryIndex =0x%08x\n"), pageDirectoryIndex));
	DBGPRINT((_L("  pageTableIndex     =0x%08x\n"), pageTableIndex));
	DBGPRINT((_L("  offsetInPage       =0x%08x\n"), offsetInPage));


	TUint32 aLevelOne = 0x41000000 + (pageDirectoryIndex << 2);
	DBGPRINT((_L("Level One address %08x\n"), aLevelOne));
	
	aLevelOne = ArloExamine(aArlo, aLevelOne);
	DBGPRINT((_L("Level One descriptor %08x\n"), aLevelOne));
	
	// we now have the level one descriptor
	if ((aLevelOne & 0x3) == 0x2) 
	{
		DBGPRINT((_L("Section\n")));
		// a Section descriptor
		return ((aLevelOne & 0xfff00000) | (aLinearAdr & 0x000fffff));
	}
	if ((aLevelOne & 0x3) != 0x1) 
	{
		DBGPRINT((_L("fault or reserved\n")));
		// Fault or reserved
		return 0;
	}


	TUint32	pageId = 0x41080000 + (pageDirectoryIndex << 2);
	pageId = ArloExamine(aArlo, pageId);
	DBGPRINT((_L("  page id=0x%08x\n"), pageId));

	TUint32		page = 0x42000000 + ((pageId >> 16) * 1024) + (pageTableIndex << 2);
//	TUint32		page = 0x42000000 + ((pageDirectoryIndex & 0xffff) * 1024) + (pageTableIndex << 2);
	page = ArloExamine(aArlo, page);
	DBGPRINT((_L("  page address=0x%08x\n"), page));

	TUint32		physAddr = (page & 0xfffff000) +  offsetInPage;
	DBGPRINT((_L("physical address=0x%08x\n"), physAddr));

	DBGPRESSKEY((_L("Linear2Physical done, press a key\n")));

	return physAddr;
}

#endif

static void ArloVirtualToPhysical(TUint32 aVirtualAddress, RArLo& aArlo)
{
	TUint32		physicalAddress;
	
	sConsole->Printf(_L("ArloVirtualToPhysical: aVirtualAddress=0x%X\n"), aVirtualAddress);
	physicalAddress = Linear2Physical(aVirtualAddress, aArlo);
	sConsole->Printf(_L("ArloVirtualToPhysical: physicalAddress=0x%X\n"), physicalAddress);
}


static TInt LoadArloDriver(RArLo& aArlo)
{
	sConsole->Printf(_L("Loading Arlo device driver\n"));

	TInt err = User::LoadLogicalDevice(KArloDriverName);
	if ((err!=KErrNone) && (err!=KErrAlreadyExists))
	{
		sConsole->Printf(_L("Error %d while loading logical device\n"), err);
	}
	else
	{
		if (err == KErrAlreadyExists)
		{
			sConsole->Printf(_L("  (driver was already loaded)\n"));
		}
		sConsole->Printf(_L("Opening the channel\n"));
		err = aArlo.Open(aArlo.VersionRequired());
		if (err) 
		{
			sConsole->Printf(_L("Error %d while opening channel to Arlo device driver\n"), err);
			User::FreeLogicalDevice(KArloDriverName);
		}
	}

	return err;
}


static void doTheArloThingL(RArLo& aArlo)
{
	TBuf<80> cline;

	sConsole->Printf(_L("Commands are:\n"));
	sConsole->Printf(_L("p <hex>      : convert virtual to physical address\n"));
//	sConsole->Printf(_L("v <hex>      : convert physical to virtual address (may crash)\n"));
	sConsole->Printf(_L("e <hex>      : examine word at physical address\n"));
	sConsole->Printf(_L("l <file>     : load file\n"));
//	sConsole->Printf(_L("o <options>  : set boot options\n"));
	sConsole->Printf(_L("c            : copy file (e.g., z:\\system\\libs\\ekern.exe) to c:\\ \n"));
	sConsole->Printf(_L("b            : boot\n"));
	sConsole->Printf(_L("x            : exit\n"));
  
  
	TUint32		aVal;
	TBool		finished = EFalse;
	do 
	{
		sCommandLine->Edit(_L("> "), &cline);

		TLex			lex(cline);
		TPtrC command = lex.NextToken();
		lex.SkipSpace();

		if (command == _L("p"))
		{
			if (lex.Val(aVal, EHex) != KErrNone) 
			{
				sConsole->Printf(_L("Illegal number\n"));
				continue;
			}

			ArloVirtualToPhysical(aVal, aArlo);
		}
		else if (command == _L("e"))
		{
			if (lex.Val(aVal, EHex) != KErrNone) 
			{
				sConsole->Printf(_L("Illegal number\n"));
				continue;
			}

			ArloExamine(aArlo, aVal);
		}
		else if (command == _L("c"))
		{
			TPtrC	fileName = lex.NextToken();
		
			ArloCopyFile(fileName);
		}
		else if (command == _L("l"))
		{
			TPtrC	fileName = lex.NextToken();
		
			LoadFile(fileName, aArlo);
		}
		else if (command == _L("b"))
		{
			ArloBoot(aArlo);
		}
		else if (command == _L("x"))
		{
			finished = ETrue;
		}
		else
		{
			sConsole->Printf(_L("Unknown command\n"));
		}
	} while (!finished);

	sConsole->Printf(_L("Bye bye....\n"));
}


static void ArloMain1L()
{
	RArLo	arlo;
	TUint	testVar = 0xdeadcafe;

	sCommandLine = CLineEdit::NewL(sConsole, 10);
	CleanupStack::PushL(sCommandLine);

	User::LeaveIfError(sFs.Connect());

	User::LeaveIfError(LoadArloDriver(arlo));

	sConsole->Printf(_L("Address of test variable=0x%x\n"), &testVar);

	TUint32 phys = RMmu::PhysicalAddress((TUint32) &testVar);
	sConsole->Printf(_L("Physical Address of test variable=0x%x\n"), phys);

	SetupArloPageTableL(arlo);

	doTheArloThingL(arlo);

	arlo.Close();
	User::FreeLogicalDevice(KArloDriverName);

	sFs.Close();

	CleanupStack::PopAndDestroy();		// sCommandLine
}


void ArloMainL()
{
	sConsole = Console::NewL(_L("ARLO"), TSize(KDefaultConsWidth,KDefaultConsHeight));
	CleanupStack::PushL(sConsole);

	TRAPD(err, ArloMain1L());
	if (err)
	{
		sConsole->Printf(_L("Oops, something failed: leave code=%d\n"), err);
		sConsole->Printf(_L("Press a key\n"));
		sConsole->Getch();
	}

	CleanupStack::PopAndDestroy();		// sConsole
}


GLDEF_C TInt E32Main() // main function called by E32
{
	__UHEAP_MARK;

	CTrapCleanup* cleanup=CTrapCleanup::New();       

	TRAPD(error, ArloMainL());                     

	__ASSERT_ALWAYS(!error, User::Panic(_L("LINUX"), error));

	delete cleanup; // destroy clean-up stack

	__UHEAP_MARKEND;

	return 0; // and return
}

